/*
 * General NeoSoft Extension Routines
 *
 *
 * $Id: neoXgeneral.c,v 1.4 1998/01/22 21:06:58 kunkee Exp $
 *
 */

#include "neo.h"
#include "util_md5.h"


/*
 *----------------------------------------------------------------------
 *
 * Neo_Incr0ObjCmd --
 *
 *	This procedure is invoked to process the "incr0" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
int
Neo_Incr0ObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    Tcl_Obj   *CONST objv[];
{
    int           value;
    Tcl_Obj      *objectPtr;
    int           increment;

    if ((objc != 2) && (objc != 3))
	return TclX_WrongArgs (interp, objv [0], " varName ?increment?\"");

    /* Get the increment amount.  If it wasn't explicitly specified, set it
     * to 1 */
    if (objc == 3) {
	if (Tcl_GetIntFromObj (interp, objv [2], &increment) != TCL_OK)
	    return TCL_ERROR;
    } else {
	increment = 1;
    }

    /* Does the variable already exist? */
    objectPtr = Tcl_ObjGetVar2 (interp, objv [1], NULL, TCL_PARSE_PART1);
    if (objectPtr == NULL) {

	/* No, create it with a default value of zero plus the increment.
	 * (in other words, the increment.) */

	Tcl_Obj *returnObj;
	Tcl_Obj *incrObj = Tcl_NewIntObj (increment);

	if ((returnObj = 
	    Tcl_ObjSetVar2 (interp, objv [1], (Tcl_Obj *) NULL, incrObj,
	      TCL_LEAVE_ERR_MSG | TCL_PARSE_PART1)) == NULL)
		return TCL_ERROR;

	Tcl_SetObjResult (interp, returnObj);

	return TCL_OK;
    }

    /* It did exist, get the integer value, increment, and set the result */
    if (Tcl_GetIntFromObj (interp, objectPtr, &value) != TCL_OK) 
	return TCL_ERROR;

    if (Tcl_IsShared(objectPtr)) {
	objectPtr = Tcl_DuplicateObj(objectPtr);
    }

    value += increment;
    Tcl_SetIntObj (objectPtr, value);
    Tcl_SetObjResult (interp, objectPtr);
    return TCL_OK; 
}

int
Neo_MD5ObjCmd (clientData, interp, objc, objv)
    ClientData  clientData;
    Tcl_Interp *interp;
    int         objc;
    Tcl_Obj   **objv;
{
    char *digest;
 
    if (objc != 2) {
        Tcl_SetResult(interp, "usage: md5 string", TCL_STATIC);
        return TCL_ERROR;
    }
    digest = md5(Tcl_GetStringFromObj(objv[1], (int*)NULL));
    Tcl_SetObjResult(interp, Tcl_NewStringObj(digest, -1));
    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * Neo_SetdefaultObjCmd --
 *
 *	This procedure is invoked to process the "setdefault" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
int
Neo_SetdefaultObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    Tcl_Obj   *CONST objv[];
{
    Tcl_Obj *objectPtr;

    if (objc != 3)
	return TclX_WrongArgs (interp, objv [0], 
		" varName value");

    /* If the variable already exists, we are done. */
    objectPtr = Tcl_ObjGetVar2 (interp, objv [1], NULL, 0);
    if (objectPtr != NULL) return TCL_OK;

    /* It didn't, create the variable and make it contain the value passed */
    if (Tcl_ObjSetVar2 (interp, objv [1], (Tcl_Obj *) NULL, objv [2],
	    TCL_LEAVE_ERR_MSG | TCL_PARSE_PART1) == NULL)
		return TCL_ERROR;
    Tcl_IncrRefCount (objv [2]);
    return TCL_OK;
}


/* 
 * Find the next unique element of a sortedlist.
 *
 * Side Effects:  index is updated.
 */
Tcl_Obj *Neo_nextUniqueElement (listObjPtrPtr, indexPtr, nElements)
Tcl_Obj         **listObjPtrPtr;
int             *indexPtr;
int              nElements;
{
    char        *thisString;
    char        *lastString;

    int          thisStringLength;
    int          lastStringLength;

    /* It's the first element of the list, an easy case as
     * we can't have seen this element already, so for sure
     * it's a winner (tho' matching dups may follow)
     */
    if (*indexPtr == 0 && nElements > 0) {
	return listObjPtrPtr[*indexPtr++];
    }

    while (1) {
	/* Nothing (or nothing more) in list?  We're done.  Return NULL */
	if (*indexPtr >= nElements) {
	    return NULL;
	}

	/* It's not the first element.  If it doesn't match the preceding
	 * one, let it fly.
	 */
	thisString = Tcl_GetStringFromObj (listObjPtrPtr[*indexPtr], &thisStringLength);
	lastString = Tcl_GetStringFromObj (listObjPtrPtr[*indexPtr - 1], &lastStringLength);

	if (!STREQU (thisString, lastString)) {
	    return listObjPtrPtr[*indexPtr++];
	}

	/* It did match the preceding one, just increment the pointer.
	 */
	indexPtr++;
    }
    /* NOTREACHED */
}


/*
 *----------------------------------------------------------------------
 *
 * Neo_Intersect3ObjCmd --
 *
 *	This procedure is invoked to process the "intersect3" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
int
Neo_Intersect3ObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    Tcl_Obj   *CONST objv[];
{
    Tcl_Obj        *inFirstList;
    Tcl_Obj        *inBothLists;
    Tcl_Obj        *inLastList;

    Tcl_Obj       **firstListObjv;
    int             firstListObjc;

    Tcl_Obj       **lastListObjv;
    int             lastListObjc;

    char           *firstString;
    char           *lastString;

    int             firstStringLength;
    int             lastStringLength;

    int             firstIndex;
    int             lastIndex;

    Tcl_Obj        *firstListElement;
    Tcl_Obj        *lastListElement;

    if (objc != 3)
	return TclX_WrongArgs (interp, objv [0], 
		" list1 list2");

    /*  */
    if (Tcl_ListObjGetElements (interp, objv [1], &firstListObjc, &firstListObjv) == TCL_ERROR) {
	return TCL_ERROR;
    }
	
    if (Tcl_ListObjGetElements (interp, objv [2], &lastListObjc, &lastListObjv) == TCL_ERROR) {
	return TCL_ERROR;
    }
	
    inFirstList = Tcl_NewListObj (0, NULL);
    inBothLists = Tcl_NewListObj (0, NULL);
    inLastList = Tcl_NewListObj (0, NULL);

    /* the code below encompasses about 90% of the cobol code
     * ever written...
     */
    firstIndex = 0;
    lastIndex = 0;

    firstListElement = Neo_nextUniqueElement (firstListObjv, &firstIndex, firstListObjc);
    lastListElement = Neo_nextUniqueElement (lastListObjv, &lastIndex, lastListObjc);
    while (1) {
	/* Is the first list done? */
	if (firstListElement == NULL) {

	    /* Yes, is the second list done? */
	    if (lastListElement == NULL) {
		/* Yes, the merge has completed. */
		break;
	    }

	    /* First list is done, but the second one is not...
	     * Copy the remainder of the second list to
	     * the inLastList list, and then we're done.
	     */
	     do {
		 Tcl_ListObjAppendElement (interp, inLastList, lastListElement);
	     } while ((lastListElement = Neo_nextUniqueElement (lastListObjv, &lastIndex, lastListObjc)) != NULL);
	     break;
	}

	/* Is the second list done? */
	if (lastListElement == NULL) {
	    /* Second list is done, but the first one is not...
	     * Copy the remainder of the first list to the
	     * inFirstList list, and then we're done.
	     */
	     do {
		 Tcl_ListObjAppendElement (interp, inFirstList, firstListElement);
	     } while ((firstListElement = Neo_nextUniqueElement (firstListObjv, &firstIndex, firstListObjc)) != NULL);
	     break;
	}

	/* Neither the first nor the second lists are done.
	 */
	firstString = Tcl_GetStringFromObj (firstListElement, &firstStringLength);
	lastString = Tcl_GetStringFromObj (lastListElement, &lastStringLength);
	switch (strcmp (firstString, lastString)) {
	    /* in first but not second */
	    case -1:
		Tcl_ListObjAppendElement (interp, inFirstList, firstListElement);
		firstListElement = Neo_nextUniqueElement (firstListObjv, &firstIndex, firstListObjc);
		break;

            /* in both */
	    case 0:
		Tcl_ListObjAppendElement (interp, inBothLists, firstListElement);
		firstListElement = Neo_nextUniqueElement (firstListObjv, &firstIndex, firstListObjc);
		lastListElement = Neo_nextUniqueElement (lastListObjv, &lastIndex, lastListObjc);
		break;

	    case 1:
		Tcl_ListObjAppendElement (interp, inLastList, lastListElement);
		lastListElement = Neo_nextUniqueElement (lastListObjv, &lastIndex, lastListObjc);
		break;
	}

    }

    return TCL_OK;
}

int
Neo_initGeneral (interp)
Tcl_Interp	*interp;
{
    Tcl_CreateObjCommand (interp, "md5",
	Neo_MD5ObjCmd, (ClientData)0,
	(Tcl_CmdDeleteProc*)NULL);

    Tcl_CreateObjCommand (interp, "incr0",
	Neo_Incr0ObjCmd, (ClientData)0, 
	(Tcl_CmdDeleteProc*) NULL);

    Tcl_CreateObjCommand (interp, "setdefault",
	Neo_SetdefaultObjCmd, (ClientData)0, 
	(Tcl_CmdDeleteProc*) NULL);

    Tcl_CreateObjCommand (interp, "_intersect3",
	Neo_Intersect3ObjCmd, (ClientData)0, 
	(Tcl_CmdDeleteProc*) NULL);

    return TCL_OK;
}
